/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "nntpgrab_automation.h"
#include "nntpgrab_utils.h"
#include "marshalers.h"
#include <glib.h>
#include <glib/gstdio.h>
#include <glib/gi18n.h>

#if GLIB_CHECK_VERSION(2,16,0)
#include <gio/gio.h>

struct _cb_data {
    GHashTable *queued_files;
};

struct _queued_file {
    time_t stamp_last_seen;
};

static gboolean process_queue(gpointer data);

#define NNTPGRAB_TYPE_AUTO_IMPORT              (auto_import_get_type ())
#define NNTPGRAB_AUTO_IMPORT(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), NNTPGRAB_TYPE_AUTO_IMPORT, NNTPGrabAutoImport))
#define NNTPGRAB_AUTO_IMPORT_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), NNTPGRAB_TYPE_AUTO_IMPORT, NNTPGrabAutoImportClass))
#define IS_NNTPGRAB_AUTO_IMPORT(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), NNTPGRAB_TYPE_AUTO_IMPORT))
#define IS_NNTPGRAB_AUTO_IMPORT_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), NNTPGRAB_TYPE_AUTO_IMPORT))
#define NNTPGRAB_AUTO_IMPORT_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), NNTPGRAB_TYPE_AUTO_IMPORT, NNTPGrabAutoImportClass))

struct NNTPGrabAutoImport
{
    GObject parent;

    GFileMonitor *monitor;
    GHashTable *queued_files;

    NGConfigOpts (*config_get_opts_func) (gpointer);
    void (*import_nzb_file_func) (gpointer, NNTPGrabNZB *, const char *, char **);
    gpointer user_data;
};

struct NNTPGrabAutoImportClass
{
    GObjectClass parent;
};

enum
{
    NZB_IMPORTED_SIGNAL,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(NNTPGrabAutoImport, auto_import, G_TYPE_OBJECT)

static void
auto_import_init (NNTPGrabAutoImport *obj)
{
    obj->monitor = NULL;
    obj->queued_files = NULL;
}

static void
auto_import_finalize (GObject *obj)
{
    NNTPGrabAutoImport *auto_import = NNTPGRAB_AUTO_IMPORT(obj);

    nntpgrab_automation_disable_auto_import(auto_import);
}

static void
auto_import_class_init (NNTPGrabAutoImportClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = auto_import_finalize;

    /**
     * NNTPGrabAutoImport::nzb-imported:
     * @obj:                An instance of the NNTPGrabAutoImport
     * @filename:           The filename of the NZB file which has just been imported
     * @collection_name:    The name of the collection in which this NZB file has been imported
     *
     * A NZB file has successfully been detected and imported to the download queue
     */
    signals[NZB_IMPORTED_SIGNAL] =  g_signal_new("nzb_imported",
                                                 G_OBJECT_CLASS_TYPE (klass),
                                                 G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                 0,
                                                 NULL, NULL,
                                                 nntpgrab_marshal_VOID__STRING_STRING,
                                                 G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);
}

static void
directory_monitor_changed (GFileMonitor *monitor, GFile *file, GFile *other_file, GFileMonitorEvent event_type, gpointer user_data)
{
    NNTPGrabAutoImport *auto_import = NNTPGRAB_AUTO_IMPORT(user_data);
    char *filename;
    struct _queued_file *queued_file;

    if (event_type != G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT &&
        event_type != G_FILE_MONITOR_EVENT_CREATED &&
        event_type != G_FILE_MONITOR_EVENT_CHANGED &&
        auto_import->queued_files != NULL) {

        return;
    }

    // only handle .nzb files
    filename = g_file_get_path(file);

    g_return_if_fail(filename != NULL);

    if (strlen(filename) < 3 || g_strncasecmp(filename + strlen(filename) - 4, ".nzb", 3)) {
        g_print("File '%s' detected by the NZB auto import, but ignored as it has the wrong extension\n", filename);
        g_free(filename);
        return;
    }

    // Have we already seen this file recently?
    queued_file = g_hash_table_lookup(auto_import->queued_files, filename);
    if (!queued_file) {
        queued_file = g_slice_alloc0(sizeof(struct _queued_file));
        g_hash_table_insert(auto_import->queued_files, filename, queued_file);

        if (g_hash_table_size(auto_import->queued_files) == 1) {
            g_timeout_add_seconds(1, process_queue, user_data);
        }
    }

    queued_file->stamp_last_seen = time(NULL);
}

static void
process_file(const char *filename, gpointer user_data)
{
    NNTPGrabAutoImport *auto_import = NNTPGRAB_AUTO_IMPORT(user_data);
    char *contents = NULL;
    GError *err = NULL;
    NNTPGrabNZB *nzbfile;
    char *errmsg = NULL;
    NGConfigOpts opts;
    char *filename_stripped;

    if (!g_file_get_contents(filename, &contents, NULL, &err)) {
        g_warning(__FILE__ ":%i Error occured while reading file\n%s:\n%s", __LINE__, filename, err->message);
        g_error_free(err);
        return;
    }

    nzbfile = nntpgrab_utils_parse_nzb_file(contents, &errmsg);
    if (!nzbfile) {
        if (!errmsg) {
            errmsg = g_strdup_printf(_("Error occured while parsing NZB File '%s'\n"), filename);
        }

        g_warning("%s", errmsg);

        g_free(errmsg);
        g_free(contents);

        return;
    }

    filename_stripped = g_path_get_basename(filename);
    nntpgrab_utils_strip_nzb_extension(filename_stripped);
    nntpgrab_utils_sanitize_collection_name(filename_stripped);

    auto_import->import_nzb_file_func(auto_import->user_data, nzbfile, filename_stripped, &errmsg);
    if (errmsg) {
        g_warning("%s", errmsg);
        g_free(errmsg);
    }
    g_free(contents);

    opts = auto_import->config_get_opts_func(auto_import->user_data);
    if (opts.move_file_after_auto_import) {
        char *basename_orig;
        char *dirname_orig;
        char *dirname;
        char *path;

        dirname_orig = g_path_get_dirname(filename);
        dirname = g_build_filename(dirname_orig, "Imported", NULL);
        g_mkdir_with_parents(dirname, 0700);

        basename_orig = g_path_get_basename(filename);
        path = g_build_filename(dirname, basename_orig, NULL);

        g_rename(filename, path);

        g_free(path);
        g_free(dirname);
        g_free(dirname_orig);
        g_free(basename_orig);
    } else {
        g_unlink(filename);
    }

    g_signal_emit_by_name(auto_import, "nzb_imported", filename, filename_stripped);
    g_free(filename_stripped);
}

static gboolean
process_queue(gpointer data)
{
    NNTPGrabAutoImport *auto_import = NNTPGRAB_AUTO_IMPORT(data);
    GHashTableIter iter;
    gpointer key, value;
    static int times_been_here_without_processing = 0;
    gboolean processed_something = FALSE;

    if (!auto_import->queued_files || times_been_here_without_processing > 10) {
        // We need to stop monitoring
        times_been_here_without_processing = 0;
        return FALSE;
    }

    g_hash_table_iter_init (&iter, auto_import->queued_files);
    while (g_hash_table_iter_next (&iter, &key, &value)) {
        struct _queued_file *queued_file = (struct _queued_file *) value;

        // Only process files which haven't changed in the last 3 seconds
        if (queued_file->stamp_last_seen + 3 > time(NULL)) {
            continue;
        }

        process_file((const char *) key, data);
        g_hash_table_iter_remove(&iter);
        processed_something = TRUE;
    }

    if (processed_something) {
        times_been_here_without_processing = 0;
    } else {
        times_been_here_without_processing++;
    }

    return TRUE;
}

static void
queued_file_free(gpointer data)
{
    g_slice_free(struct _queued_file, data);
}

NGType
nntpgrab_automation_auto_import_get_type(void)
{
    return (NGType) auto_import_get_type();
}

NNTPGrabAutoImport *
nntpgrab_automation_auto_import_new(NGConfigOpts (*config_get_opts_func) (void *user_data), void (*import_nzb_file_func) (void *user_data, NNTPGrabNZB *nzbfile, const char *filename, char **errmsg), void *user_data)
{
    NNTPGrabAutoImport *auto_import;

    g_return_val_if_fail(config_get_opts_func != NULL, NULL);
    g_return_val_if_fail(import_nzb_file_func != NULL, NULL);

    auto_import = g_object_new(NNTPGRAB_TYPE_AUTO_IMPORT, NULL);
    auto_import->config_get_opts_func = config_get_opts_func;
    auto_import->import_nzb_file_func = import_nzb_file_func;
    auto_import->user_data = user_data;

    return auto_import;
}

void
nntpgrab_automation_auto_import_destroy(NNTPGrabAutoImport *auto_import)
{
    g_return_if_fail(auto_import != NULL);

    g_object_unref(auto_import);
}

ngboolean
nntpgrab_automation_enable_auto_import(NNTPGrabAutoImport *auto_import, const char *directory, char **errmsg)
{
    g_return_val_if_fail(auto_import != NULL, FALSE);
    g_return_val_if_fail(directory != NULL, FALSE);

    // Just to be sure
    nntpgrab_automation_disable_auto_import(auto_import);

    auto_import->queued_files = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, queued_file_free);

    auto_import->monitor = nntpgrab_utils_monitor_directory(directory, errmsg);
    if (!auto_import->monitor) {
        return FALSE;
    }

    g_signal_connect(auto_import->monitor, "changed", G_CALLBACK(directory_monitor_changed), auto_import);

    return TRUE;
}

void
nntpgrab_automation_disable_auto_import(NNTPGrabAutoImport *auto_import)
{
    g_return_if_fail(auto_import != NULL);

    if (auto_import->monitor) {
        nntpgrab_utils_cancel_directory_monitor(auto_import->monitor);
        auto_import->monitor = NULL;
        g_hash_table_unref(auto_import->queued_files);
        auto_import->queued_files = NULL;
    }
}

#endif
