/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include <glib-object.h>
#include <libintl.h>

#include "nntpgrab_utils.h"

#ifdef HAVE_SOUP
#include <libsoup/soup.h>
#endif
#ifdef HAVE_SOUP_GNOME
#include <libsoup/soup-gnome.h>
#endif
#if defined(WIN32) && GLIB_CHECK_VERSION(2,27,1)
#include <glib/gstdio.h>
#include <shlobj.h>
#endif
#if defined(DARWIN)
#include <CoreFoundation/CoreFoundation.h>
#endif

NGTypeInstance*   ng_type_check_instance_cast   (NGTypeInstance      *instance,
                                                 NGType               iface_type)
{
    return (NGTypeInstance*) g_type_check_instance_cast((GTypeInstance*) instance, (GType) iface_type);
}

#if defined(WIN32) && GLIB_CHECK_VERSION(2,27,1)
static gchar *
get_special_folder (int csidl)
{
    wchar_t path[MAX_PATH+1];
    HRESULT hr;
    LPITEMIDLIST pidl = NULL;
    BOOL b;
    gchar *retval = NULL;

    hr = SHGetSpecialFolderLocation (NULL, csidl, &pidl);
    if (hr == S_OK) {
        b = SHGetPathFromIDListW (pidl, path);
        if (b)
            retval = g_utf16_to_utf8 (path, -1, NULL, NULL, NULL);
        CoTaskMemFree (pidl);
    }
    return retval;
}

static void
move_win32_configuration_files_if_necessary(void)
{
    char *appdata_dir = get_special_folder (CSIDL_APPDATA);
    char *config_dir = g_build_path(G_DIR_SEPARATOR_S, appdata_dir, "NNTPGrab", NULL);

    if (g_file_test(config_dir, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) {
        char *config_dir_new = g_build_path(G_DIR_SEPARATOR_S, g_get_user_config_dir(), "NNTPGrab", NULL);
        g_rename(config_dir, config_dir_new);
        g_free(config_dir_new);
    }

    g_free(config_dir);
    g_free(appdata_dir);
}
#endif

void nntpgrab_utils_perform_base_initialization(void)
{
#if defined(WIN32) || defined(DARWIN)
    char *dirname;
#endif
#if defined(DARWIN)
    CFURLRef appUrlRef;
    CFStringRef macPath;
    const char *pathPtr;
#elif defined(WIN32)
    char *dirname2;
#endif

    if (!g_thread_supported ()) g_thread_init (NULL);
    g_type_init();

#ifdef DARWIN
     appUrlRef = CFBundleCopyBundleURL(CFBundleGetMainBundle());
     macPath = CFURLCopyFileSystemPath(appUrlRef, kCFURLPOSIXPathStyle);
     pathPtr = CFStringGetCStringPtr(macPath, CFStringGetSystemEncoding());
     /* Hack for the Gtk frontend which is started a bit different than the Qt frontend */
     if (strstr(pathPtr, "Contents/Resources/bin")) {
         dirname = g_build_path(G_DIR_SEPARATOR_S, pathPtr, "..", "share", "locale", NULL);
     } else {
         dirname = g_build_path(G_DIR_SEPARATOR_S, pathPtr, "Contents", "Resources", "share", "locale", NULL);
     }
     bindtextdomain (GETTEXT_PACKAGE, dirname);
     g_free(dirname);
     CFRelease(appUrlRef);
     CFRelease(macPath);
#elif defined(WIN32)
    dirname = g_win32_get_package_installation_directory_of_module(NULL);
    dirname2 = g_build_path(G_DIR_SEPARATOR_S, dirname, "\\share\\locale", NULL);
    bindtextdomain (GETTEXT_PACKAGE, dirname2);
    g_free(dirname2);
#else
    bindtextdomain (GETTEXT_PACKAGE, NNTPGRABLOCALEDIR);
#endif
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
    textdomain (GETTEXT_PACKAGE);

#ifdef HAVE_SOUP
    /* Due to a bug in libsoup we need to make sure that all GObject's
     * are initialized from the main thread to avoid a deadlock when libsoup
     * is called for the first time from a different thread */
    SoupSession *session = soup_session_sync_new();

#if 0
    SoupLogger *logger = soup_logger_new(SOUP_LOGGER_LOG_BODY, -1);
    soup_session_add_feature(session, SOUP_SESSION_FEATURE(logger));
    g_object_unref(logger);
#endif

#ifdef HAVE_SOUP_GNOME
    /* Enable proxy support when using GNOME 2.26.0 or higher */
    soup_session_add_feature_by_type(session, SOUP_TYPE_GNOME_FEATURES_2_26);
#endif

#ifdef HAVE_SOUP_GZIP
    /* Enable gzip compression */
    soup_session_add_feature_by_type(session, SOUP_TYPE_CONTENT_DECODER);
#endif

    g_object_unref(session);
#endif

#if defined(WIN32) && GLIB_CHECK_VERSION(2,27,1)
    /* As of GLib 2.27.1 the return value of the function g_get_user_data_dir()
     * was changed to return CSIDL_LOCAL_APPDATA instead of CSIDL_APPDATA on Win32.
     * To prevent a broken NNTPGrab configuration, check if there are configuration
     * files at the old path and move them to the new location if necessary */
    move_win32_configuration_files_if_necessary();
#endif
}
