/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_GIO_SOCKET

#include <string.h>
#include <gio/gio.h>

#include "nntpconnection_backend_gio.h"
#include "nntpgrab_plugin.h"

struct _NntpconnectionBackendGioPrivate {
    GMainLoop *loop;
    GSocketConnectable *connectable;
    GSocket *socket;

    GIOStream *connection;
    GDataInputStream *istream;
    GOutputStream *ostream;
};

struct _NntpconnectionBackendGioClass {
    GObjectClass parent_class;
    /* define vtable methods and signals here */
};

static void nntpconnection_backend_gio_disconnect(NntpConnection *conn);

static void nntpconnection_backend_gio_iface_init(NntpconnectionBackendIface *iface);
static void nntpconnection_backend_gio_init (NntpconnectionBackendGio *self)
{
    self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self, NNTPCONNECTION_TYPE_BACKEND_GIO, NntpconnectionBackendGioPrivate);

    self->priv->loop = g_main_loop_new(NULL, FALSE);
}

static void nntpconnection_backend_gio_class_init (NntpconnectionBackendGioClass *klass)
{
    g_type_class_add_private (klass, sizeof (NntpconnectionBackendGioPrivate));
}

G_DEFINE_TYPE_WITH_CODE (NntpconnectionBackendGio, nntpconnection_backend_gio, NNTP_TYPE_CONNECTION,
                         G_IMPLEMENT_INTERFACE(NNTPCONNECTION_TYPE_BACKEND, nntpconnection_backend_gio_iface_init))

NntpConnection *nntpconnection_backend_gio_new (void)
{
    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Created NNTP connection using backend 'gio'");
    return NNTP_CONNECTION(g_object_new(NNTPCONNECTION_TYPE_BACKEND_GIO, NULL));
}

static NNTPConnectionErrCode
nntpconnection_backend_gio_connect(NntpConnection *conn, const char *hostname, int port, char **errmsg)
{
    NntpconnectionBackendGioPrivate *conn_priv = NNTPCONNECTION_BACKEND_GIO(conn)->priv;
    GError *err = NULL;
    GSocketAddressEnumerator *enumerator;
    GSocketAddress *address;
    GCancellable *cancellable = NULL;

    g_return_val_if_fail(conn != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(conn->poll_fd.fd == -1, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(hostname != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(port > 0, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(conn_priv != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(errmsg != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(*errmsg == NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);

    conn_priv->socket = g_socket_new(G_SOCKET_FAMILY_IPV4, G_SOCKET_TYPE_STREAM, 0, &err);
    if (!conn_priv->socket) {
        *errmsg = g_strdup_printf("g_socket_new() FAILED: %s", err->message);
        g_error_free(err);
        return NNTP_CONNECTION_ERROR_INVALID_ARGUMENT;
    }

    g_socket_set_timeout(conn_priv->socket, 30);

    conn_priv->connectable = g_network_address_parse(hostname, port, &err);
    if (!conn_priv->connectable) {
        *errmsg = g_strdup_printf("Unable to parse hostname '%s' and port '%i': %s", hostname, port, err->message);
        g_error_free(err);
        nntpconnection_backend_gio_disconnect(conn);
        return NNTP_CONNECTION_ERROR_HOST_NOT_FOUND;
    }

    enumerator = g_socket_connectable_enumerate(conn_priv->connectable);
    while (TRUE) {
        address = g_socket_address_enumerator_next (enumerator, cancellable, &err);
        if (!address) {
#if 0
            if (err == NULL)
                g_printerr ("%s: No more addresses to try\n", argv[0]);
            else
                g_printerr ("%s: %s\n", argv[0], error->message);
#endif
            return 1;
        }

        if (g_socket_connect(conn_priv->socket, address, cancellable, &err))
            break;

        //g_printerr ("%s: Connection to %s failed: %s, trying next\n", argv[0], socket_address_to_string (address), err->message);
        g_error_free (err);
        err = NULL;

        g_object_unref (address);
    }

    g_object_unref (enumerator);

    /* Keep the GSocketConnectable around if we still need to set up SSL later */
    if (!conn->server_info.use_ssl) {
        g_object_unref(conn_priv->connectable);
        conn_priv->connectable = NULL;
    }

    conn_priv->connection = G_IO_STREAM (g_socket_connection_factory_create_connection(conn_priv->socket));
    conn_priv->istream = g_data_input_stream_new(g_io_stream_get_input_stream(conn_priv->connection));
    conn_priv->ostream = g_io_stream_get_output_stream(conn_priv->connection);

    g_socket_set_blocking(conn_priv->socket, FALSE);
    g_data_input_stream_set_newline_type(G_DATA_INPUT_STREAM(conn_priv->istream), G_DATA_STREAM_NEWLINE_TYPE_CR_LF);

    conn->poll_fd.fd = g_socket_get_fd(conn_priv->socket);

    return NNTP_CONNECTION_ERROR_NONE;
}

static void
nntpconnection_backend_gio_disconnect(NntpConnection *conn)
{
    NntpconnectionBackendGioPrivate *conn_priv = NNTPCONNECTION_BACKEND_GIO(conn)->priv;

    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn_priv != NULL);

    if (conn_priv->socket) {
        g_object_unref(conn_priv->socket);
        conn_priv->socket = NULL;
    }
}

static gboolean
accept_certificate (GTlsClientConnection *conn, GTlsCertificate *cert,
                    GTlsCertificateFlags errors, gpointer user_data)
{
    g_print ("Certificate would have been rejected ( ");
    if (errors & G_TLS_CERTIFICATE_UNKNOWN_CA)
        g_print ("unknown-ca ");
    if (errors & G_TLS_CERTIFICATE_BAD_IDENTITY)
        g_print ("bad-identity ");
    if (errors & G_TLS_CERTIFICATE_NOT_ACTIVATED)
        g_print ("not-activated ");
    if (errors & G_TLS_CERTIFICATE_EXPIRED)
        g_print ("expired ");
    if (errors & G_TLS_CERTIFICATE_REVOKED)
        g_print ("revoked ");
    if (errors & G_TLS_CERTIFICATE_INSECURE)
        g_print ("insecure ");
    g_print (") but accepting anyway.\n");

    return TRUE;
}

static gboolean
nntpconnection_backend_gio_prepare_ssl (NntpConnection *conn, char **errmsg)
{
#if GLIB_CHECK_VERSION(2,28,0)
    GError *err = NULL;
    NntpconnectionBackendGioPrivate *conn_priv;
    GIOStream *tls_conn;

    g_return_val_if_fail(conn != NULL, FALSE);
    g_return_val_if_fail(conn->server_info.use_ssl == TRUE, FALSE);
    g_return_val_if_fail(errmsg != NULL, FALSE);
    g_return_val_if_fail(*errmsg == NULL, FALSE);

    conn_priv = NNTPCONNECTION_BACKEND_GIO(conn)->priv;
    g_return_val_if_fail(conn_priv != NULL, FALSE);
    g_return_val_if_fail(conn_priv->connection != NULL, FALSE);
    g_return_val_if_fail(conn_priv->connectable != NULL, FALSE);

    tls_conn = g_tls_client_connection_new (conn_priv->connection, conn_priv->connectable, &err);
    if (!tls_conn) {
        *errmsg = g_strdup_printf("Could not create TLS connection: %s\n", err->message);
        g_error_free(err);
        return FALSE;
    }

    g_signal_connect (tls_conn, "accept-certificate", G_CALLBACK (accept_certificate), NULL);

    if (!g_tls_connection_handshake (G_TLS_CONNECTION (tls_conn), NULL, &err)) {
        *errmsg = g_strdup_printf("Error during TLS handshake: %s\n", err->message);
        g_error_free(err);
        return FALSE;
    }

    /* Replace the already known streams with the TLS based one */
    g_object_unref(conn_priv->connection);
    g_filter_input_stream_set_close_base_stream(G_FILTER_INPUT_STREAM(conn_priv->istream), FALSE);
    g_object_unref(conn_priv->istream);
    conn_priv->connection = G_IO_STREAM(tls_conn);
    conn_priv->istream = g_data_input_stream_new(g_io_stream_get_input_stream(conn_priv->connection));
    conn_priv->ostream = g_io_stream_get_output_stream(conn_priv->connection);

    g_data_input_stream_set_newline_type(G_DATA_INPUT_STREAM(conn_priv->istream), G_DATA_STREAM_NEWLINE_TYPE_CR_LF);

    g_object_unref(conn_priv->connectable);
    conn_priv->connectable = NULL;

    return TRUE;
#else
    *errmsg = g_strdup("This version of GIO doesn't have SSL/TLS support");
    return FALSE;
#endif
}

static gboolean
source_ready (GPollableInputStream *stream,
              gpointer              data)
{
    NntpconnectionBackendGioPrivate *conn_priv = (NntpconnectionBackendGioPrivate*) data;

    g_main_loop_quit (conn_priv->loop);

    return FALSE;
}

static void
ensure_connection_condition (NntpConnection *conn,
                             GIOCondition    condition,
                             GCancellable   *cancellable)
{
    NntpconnectionBackendGioPrivate *conn_priv = NNTPCONNECTION_BACKEND_GIO(conn)->priv;
    GSource *source;

    if (condition & G_IO_IN)
        source = g_pollable_input_stream_create_source (G_POLLABLE_INPUT_STREAM (g_io_stream_get_input_stream (conn_priv->connection)), cancellable);
    else
        source = g_pollable_output_stream_create_source (G_POLLABLE_OUTPUT_STREAM (g_io_stream_get_output_stream (conn_priv->connection)), cancellable);

    g_source_set_callback (source, (GSourceFunc) source_ready, conn_priv, NULL);
    g_source_attach (source, NULL);
    g_source_unref (source);
    g_main_loop_run (conn_priv->loop);
}

struct _cancel_data {
    GCancellable *cancellable;
    int timeout;
};

static gpointer
cancel_thread (gpointer data)
{
    struct _cancel_data *cancel_data = (struct _cancel_data*) data;

    g_usleep(cancel_data->timeout);
    g_print("Cancelling\n");
    g_cancellable_cancel(cancel_data->cancellable);

    g_slice_free1(sizeof(struct _cancel_data), cancel_data);

    return NULL;
}

static gboolean
nntpconnection_backend_gio_ensure_data_is_ready(NntpConnection *conn, int timeout)
{
    GCancellable *cancellable = NULL;
    gboolean ret;

    if (timeout > 0) {
        static GCancellable *cancellable_cached = NULL;
        struct _cancel_data *cancel_data = g_slice_alloc(sizeof(struct _cancel_data));

        if (!cancellable_cached) {
            cancellable_cached = g_cancellable_new();
        }

        cancellable = cancellable_cached;
        cancel_data->cancellable = cancellable_cached;
        cancel_data->timeout = timeout;

        g_thread_create(cancel_thread, cancel_data, FALSE, NULL);
    }

    ensure_connection_condition (conn, G_IO_IN, cancellable);

    ret = !(g_cancellable_is_cancelled(cancellable));
    g_cancellable_reset(cancellable);

    return ret;
}

static gboolean
nntpconnection_backend_gio_read_msg(NntpConnection *conn, gboolean read_line, int max_length, void *data, int *data_length, gboolean *more_data_ready, gboolean *try_again, char **errmsg)
{
    NntpconnectionBackendGioPrivate *conn_priv;
    GError *err = NULL;
    gsize len;
    char *line;

    g_return_val_if_fail(conn != NULL, FALSE);
    g_return_val_if_fail(max_length > 0, FALSE);
    g_return_val_if_fail(data != NULL, FALSE);
    g_return_val_if_fail(data_length != NULL, FALSE);
    g_return_val_if_fail(*data_length == 0, FALSE);
    /* more_data_ready can be NULL */
    g_return_val_if_fail(try_again != NULL, FALSE);
    g_return_val_if_fail(errmsg != NULL, FALSE);
    g_return_val_if_fail(*errmsg == NULL, FALSE);

    conn_priv = NNTPCONNECTION_BACKEND_GIO(conn)->priv;
    g_return_val_if_fail(conn_priv != NULL, FALSE);

    *try_again = FALSE;

    if (read_line) {
        line = g_data_input_stream_read_line(G_DATA_INPUT_STREAM(conn_priv->istream), &len, NULL, &err);
    } else {
        line = g_data_input_stream_read_upto(G_DATA_INPUT_STREAM(conn_priv->istream), "\n", -1, &len, NULL, &err);
        if (line) {
            /* Also read the \n character itself */
            g_return_val_if_fail(g_data_input_stream_read_byte(G_DATA_INPUT_STREAM(conn_priv->istream), NULL, NULL) == '\n', FALSE);
        }
    }

    if (line) {
        /* NOTE: len will include the newline characters when read_line == FALSE, otherwise len will be 0 when an empty line was received */

        if (len > max_length) {
            ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "nntpconnection_backend_gio_read_msg: Very long line received from server. Ignoring\nLine = %s", line);
            g_free(line);
            return FALSE;
        }

        memcpy(data, line, len);

        /* g_data_input_stream_read_upto doesn't put the \n character in the buffer. Manually add this when read_line == FALSE */
        if (!read_line && len + 1 < max_length) {
            char *data_ptr = (char*) data;
            data_ptr[len] = '\n';
            data_ptr[len + 1] = '\0';
            len += 1;
        }

        g_free(line);
    } else {
        len = -1;
    }

    if (len == -1) {
        g_return_val_if_fail(err != NULL, FALSE);

        if (err->code == G_IO_ERROR_WOULD_BLOCK) {
            *try_again = TRUE;
        }
        *errmsg = g_strdup(err->message);
        g_error_free(err);
        return FALSE;
    } else {
        *data_length = len;

        nntpconnection_backend_perform_shaping(conn->max_bandwidth, len);
        nntpconnection_backend_update_traffic_monitor(len);
    }

    if (more_data_ready) {
        *more_data_ready = (g_buffered_input_stream_get_available(G_BUFFERED_INPUT_STREAM(conn_priv->istream)) > 0);
    }

    return TRUE;
}

static gboolean
nntpconnection_backend_gio_send_msg(NntpConnection *conn, void *data, int length, char **errmsg)
{
    NntpconnectionBackendGioPrivate *conn_priv;
    GError *err = NULL;
    gboolean ret;

    g_return_val_if_fail(conn != NULL, FALSE);

    conn_priv = NNTPCONNECTION_BACKEND_GIO(conn)->priv;
    g_return_val_if_fail(conn_priv != NULL, FALSE);

    ret = g_output_stream_write_all(G_OUTPUT_STREAM(conn_priv->ostream), data, length, NULL, NULL, &err);
    if (!ret) {
        *errmsg = g_strdup(err->message);
        g_error_free(err);
        return FALSE;
    }

    return TRUE;
}

static gboolean
nntpconnection_backend_gio_get_has_ssl_support(NntpConnection *conn)
{
#if GLIB_CHECK_VERSION(2,28,0)
    return g_tls_backend_supports_tls(g_tls_backend_get_default());
#else
    return FALSE;
#endif
}

static void nntpconnection_backend_gio_iface_init(NntpconnectionBackendIface *iface)
{
    iface->connect = nntpconnection_backend_gio_connect;
    iface->disconnect = nntpconnection_backend_gio_disconnect;
    iface->prepare_ssl = nntpconnection_backend_gio_prepare_ssl;
    iface->ensure_data_is_ready = nntpconnection_backend_gio_ensure_data_is_ready;
    iface->read_msg = nntpconnection_backend_gio_read_msg;
    iface->send_msg = nntpconnection_backend_gio_send_msg;
    iface->get_has_ssl_support = nntpconnection_backend_gio_get_has_ssl_support;
}

#endif
